from __future__ import division, print_function, unicode_literals, absolute_import

import numpy as np

try:
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_agg import FigureCanvasAgg
except ImportError:
    FigureCanvasAgg = None

from .utility import image_data_mode, setup_uint8
from .compression import compress


def figure_to_image(fig, fmt='raw', dpi=None, close_figure=True):
    """Generate compressed-image representation of a Matplotlib figure.
    Nothing gets displayed to the screen.

    See here for partial inspiration: https://gist.github.com/rduplain/1641344
    """

    if not FigureCanvasAgg:
        raise ImportError('Matplotlib is not installed')

    if dpi:
        dpi_original = fig.get_dpi()
        fig.set_dpi(dpi)

    # Render figure to buffer as uncompressed RGBA
    canvas = FigureCanvasAgg(fig)
    buff, (num_samples, num_lines) = canvas.print_to_buffer()
    num_colors = 4

    # Grab data from buffer
    data = np.frombuffer(buff, dtype=np.uint8)
    data.shape = num_lines, num_samples, num_colors

    # Close the figure
    if close_figure:
        plt.close(fig)


    mode = image_data_mode(data)
    if mode.lower() != 'rgba':
        raise ValueError('Unexpected image mode: {:s}'.format(mode))

    if dpi:
        fig.set_dpi(dpi_original)

    if fmt == 'raw':
        # Return uncomprressed image byte data
        return data
    else:
        # Return compressed data
        return compress(data, mode=mode, fmt=fmt)



def figure_to_file(fig, fname, fmt='png', dpi=None, close_figure=True):
    """Write selected figure to image file
    """
    b, e = os.path.splitext(fname)

    if fmt:
        e = '.' + fmt.lower()
    else:
        fmt = e[1:].lower()

    f_out = b + e

    data_comp = figure_to_image(fig, fmt=fmt, dpi=dpi, close_figure=close_figure)

    with open(f_out, 'wb') as fp:
        fp.write(data_comp)

    # Done

#------------------------------------------------

if __name__ == '__main__':
    pass




from __future__ import division, print_function, unicode_literals, absolute_import

import io

import numpy as np
import PIL.Image

from . import image_io
from . import utility


def compress(data, fmt, **kwargs):
    """Compress image, return data buffer.

    fmt: 'png', 'jpeg', etc.

    Alpha channel will be ignored if fmt == 'jpeg'.

    Returns a byte buffer of compressed data.

    Parameter options: http://pillow.readthedocs.org/handbook/image-file-formats.html
    """
    # Enforce 8-byte data
    data = utility.setup_uint8(data)

    fmt = fmt.lower()
    if fmt == 'jpg':
        fmt = 'jpeg'

    if fmt == 'jpeg':
        if data.ndim == 3:
            if data.shape[2] == 4:
                # Discard alpha channel for JPEG compression.
                data = data[:, :, :2]

    buff = io.BytesIO()
    image_io.write(buff, data, fmt, **kwargs)

    data_comp = buff.getvalue()

    return data_comp



def decompress(data_comp):
    """Decompress image from supplied byte data.
    """
    buff = io.BytesIO(data_comp)
    img = PIL.Image.open(buff)

    data = np.asarray(img)

    return data

#------------------------------------------------

if __name__ == '__main__':
    pass




from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from . import utility


def _resize(data, factor, axis=0, verbose=False):
    """
    Do the work of resizing image data along horizontal or vertical axis.

    factor:
        positive value larger than one implies grow by integer amount
        positive value between zero and one implies shrink by integer amount (1/factor)
        negative value less than negative one implies shrink by integer amount abs(factor)

    data: 3D array
    """
    if axis not in [0, 1]:
        raise ValueError('Invalid axis: {}'.format(axis))

    num_lines, num_samples, num_bands = data.shape

    if factor < -1:
        factor = 1 / round(abs(factor))
    elif 0 < factor and factor < 1:
        factor = 1 / round(1/factor)
    elif 1 == factor:
        pass
    elif 1 < factor:
        factor = round(factor)
    else:
        raise ValueError('Invalid scale factor: {}'.format(factor))

    # Do nothing for unit scale factor.
    if factor == 1:
        return data

    # New sizes for new data.
    if axis == 0:
        num_lines_new = int(num_lines * factor)
        num_samples_new = num_samples
    else:
        num_lines_new = num_lines
        num_samples_new = int(num_samples * factor)

    if verbose:
        print(axis, factor)
        print(num_lines, num_samples)
        print(num_lines_new, num_samples_new)

    # Build new data.
    data_new = np.zeros((num_lines_new, num_samples_new, num_bands), dtype=np.float)

    if factor < 1:
        # Co-add along specified image axis.

        coadd = int(1/factor)
        for p in range(0, coadd):
            if axis == 0:
                data_p = data[p:num_lines_new*coadd:coadd, :, :]
            else:
                data_p = data[:, p:num_samples_new*coadd:coadd, :]

            data_new += data_p

        data_new /= coadd

    else:
        # Grow by nearest-neighbor sampling.
        for p in range(0, factor):
            if axis == 0:
                data_new[p::factor, :, :] = data
            else:
                data_new[:, p::factor, :] = data

    return data_new


def rebin(data, factor, verbose=False):
    """
    Resize image by integer amount.
    Pixel-averaging for shrinking, nearest-neighbor for growing.

    size_factor: integer or 2-tuple of X and Y integer scale factors.
    """

    if not hasattr(factor, '__iter__'):
        factor = [factor, factor]

    factor_lines, factor_samples = factor

    if factor_lines == 1 and factor_samples == 1:
        return data

    dtype_orig = data.dtype

    # Prepare the data.
    data = np.asarray(data)

    ndim_orig = data.ndim

    num_lines, num_samples = data.shape[0:2]
    data = data.reshape(num_lines, num_samples, -1)

    # Do the work.
    data_tmp = _resize(data, factor_lines, axis=0, verbose=verbose)
    data_new = _resize(data_tmp, factor_samples, axis=1, verbose=verbose)

    if ndim_orig == 2:
        data_new = data_new.squeeze()

    # Finish.
    if utility._is_integer_type(dtype_orig):
        data_new = np.round(data_new)

    data_new = data_new.astype(dtype_orig)

    # Done.
    return data_new

#------------------------------------------------

if __name__ == '__main__':
    pass



from __future__ import division, print_function, unicode_literals, absolute_import

import os
import io
import urllib

import numpy as np
import PIL.Image
import requests


def read(fp):
    """Read image data from file-like object using PIL.  Return Numpy array.
    """
    with open(fp, 'rb') as fpp:
        img = PIL.Image.open(fpp)
        data = np.asarray(img)

    return data



def write(fp, data, fmt=None, **kwargs):
    """Write image data from Numpy array to file-like object.

    File format is automatically determined from fp if it's a filename, otherwise you
    must specify format via fmt keyword, e.g. fmt = 'png'.

    Parameter options: http://pillow.readthedocs.io/en/4.2.x/handbook/image-file-formats.html
    """
    img = PIL.Image.fromarray(data)
    img.save(fp, format=fmt, **kwargs)



def normalize_url(url):
    parts = urllib.parse.urlsplit(url)
    url_nice = parts.geturl()

    return url_nice



def download(url, verbose=False):
    """Download compressed image data from URL.

    http://stackoverflow.com/questions/13137817/how-to-download-image-using-requests/13137873#13137873
    """
    url = normalize_url(url)

    resp = requests.get(url)
    if not resp.ok:
        msg = 'Problem fetching data: {}'.format(resp.reason)
        # raise requests.RequestException(msg)
        raise IOError(msg)

    # Binary compressed image data from response content
    data_comp = resp.content

    # Image compression format, e.g. 'image/jpeg' --> 'jpeg'
    fmt = resp.headers['content-type'].split('/')[1]

    # Done
    return data_comp, fmt

#------------------------------------------------

if __name__ == '__main__':
    pass




from __future__ import division, print_function, unicode_literals

"""
Test Examples
=============


## Python Tests
self.assertTrue(value)
self.assertFalse(value)

self.assertGreater(first, second, msg=None)
self.assertGreaterEqual(first, second, msg=None)
self.assertLess(first, second, msg=None)
self.assertLessEqual(first, second, msg=None)

self.assertAlmostEqual(first, second, places=7, msg=None, delta=None)
self.assertNotAlmostEqual(first, second, places=7, msg=None, delta=None)

self.assertItemsEqual(actual, expected, msg=None)
self.assertSequenceEqual(seq1, seq2, msg=None, seq_type=None)
self.assertListEqual(list1, list2, msg=None)
self.assertTupleEqual(tuple1, tuple2, msg=None)
self.assertSetEqual(set1, set2, msg=None)
self.assertDictEqual(expected, actual, msg=None)

self.assertRaises(Exception, some_func, arg, arg_nother)


## Numpy Tests
np.testing...

assert_equal(actual, desired[, err_msg, verbose])   Raises an AssertionError if two objects are not equal.
assert_allclose(actual, desired[, rtol, ...])       Raises an AssertionError if two objects are not equal up to desired tolerance.

assert_almost_equal(actual, desired[, ...])         Raises an AssertionError if two items are not equal up to desired precision.
assert_approx_equal(actual, desired[, ...])         Raises an AssertionError if two items are not equal up to significant digits.
assert_array_almost_equal(x, y[, decimal, ...])     Raises an AssertionError if two objects are not equal up to desired precision.
assert_array_almost_equal_nulp(x, y[, nulp])        Compare two arrays relatively to their spacing.
assert_array_max_ulp(a, b[, maxulp, dtype])         Check that all items of arrays differ in at most N Units in the Last Place.
assert_array_equal(x, y[, err_msg, verbose])        Raises an AssertionError if two array_like objects are not equal.
assert_array_less(x, y[, err_msg, verbose])         Raises an AssertionError if two array_like objects are not ordered by less than.
assert_raises(exception_class, callable, ...)       Fail unless an exception of class exception_class is thrown by callable when invoked with arguments args and keyword arguments kwargs.
assert_raises_regex(exception_class, ...)           Fail unless an exception of class exception_class and with message that matches expected_regexp is thrown by callable when invoked with arguments args and keyword arguments kwargs.
assert_warns(warning_class, *args, **kwargs)        Fail unless the given callable throws the specified warning.
assert_string_equal(actual, desired)                Test if two strings are equal.

"""

import unittest
import os
import pathlib
import shutil

import numpy as np

import context
import image_attendant as imat

_path_module = pathlib.Path(__file__).parent.absolute()


#------------------------------------------------
class TestRebin(unittest.TestCase):
    def setUp(self):
        self.path_Lena = pathlib.Path(_path_module, 'Lena')

    def tearDown(self):
        pass


    def test_compress_png(self):
        files = self.path_Lena.glob('Lena*')
        for f in files:
            img_in = imat.read(f)

            img_test = imat.rebin(img_in, 0.5)
            print(img_in.shape, img_test.shape)

            a = img_in.shape[:2]

            b1, b2 = img_test.shape[:2]
            b = b1*2, b2*2
            np.testing.assert_equal(a, b, str(f))

            if img_in.ndim == 3:
                c = img_in.shape[2]
                d = img_test.shape[2]

                np.testing.assert_equal(c, d, str(f))
#------------------------------------------------

if __name__ == '__main__':
    unittest.main(verbosity=2)



from __future__ import division, print_function, unicode_literals, absolute_import

from .image_io import read, write, download
from .compression import compress, decompress
from .mpl import figure_to_image, figure_to_file
from .rebin import rebin
from .utility import image_data_mode, setup_uint8, data_url, collapse_alpha



from __future__ import division, print_function, unicode_literals, absolute_import

import numpy as np

_dtypes_float = [np.float, np.float16, np.float32]
_dtypes_int = [np.int, np.int8, np.int16, np.int32, np.uint8, np.uint16, np.uint32]


def _is_float_type(data):
    if type(data) == np.dtype:
        dtype_data = data
    else:
        data = np.asarray(data)
        dtype_data = data.dtype

    return dtype_data in _dtypes_float


def _is_integer_type(data):
    if type(data) == np.dtype:
        dtype_data = data
    else:
        data = np.asarray(data)
        dtype_data = data.dtype

    return dtype_data in _dtypes_int

#------------------------------------------------


def image_data_mode(data):
    """Determine image color mode compatible with PIL / Pillow
    Input data is expected to be 2D or 3D: [num_lines, num_samples, num_bands].
    """
    # Force data to be Numpy ndarray, if not already.
    data = np.asarray(data)

    if data.ndim == 2:
        num_bands = 1
    elif data.ndim == 3:
        num_bands = data.shape[2]
    else:
        raise ValueError('Invalid number of data dimensions: {}'.format(data.ndim))

    if num_bands == 1:
        mode = 'L'
    elif num_bands == 3:
        mode = 'RGB'
    elif num_bands == 4:
        mode = 'RGBA'
    else:
        raise ValueError('Invalid number of bands.')

    return mode



def setup_uint8(data, lohi=None):
    """Ensure data is unsigned bytes

    If data type is not np.uint8 it will be converted by scaling
    min(data) -> 0 and max(data) -> 255.
    """
    data = np.asarray(data)

    # Scale to np.uint8?
    if not (data.dtype == np.uint8) or lohi is not None:
        data = data.astype(np.float32)

        if lohi is None:
            lohi = data.min(), data.max()

        lo, hi = lohi
        if lo == hi:
            raise ValueError('Invalid data range: {}, {}'.format(lo, hi))

        data = (data - lo) / (hi - lo)
        data = np.clip(data, 0, 1)*255
        data = np.round(data).astype(np.uint8)

    return data



def collapse_alpha(data_rgba):
    """Collapse alpha channel

    https://en.wikipedia.org/wiki/Alpha_compositing#Alpha_blending
    """
    data_rgba = np.asarray(data_rgba)

    if data_rgba.dtype != np.uint8:
        # Nothing to do
        return data_rgba

    if data_rgba.ndim != 3:
        # Nothing to do
        return data_rgba

    if data_rgba.shape[2] != 4:
        # Nothing to do
        return data_rgba

    # Convert to float, between 0 and 1.
    data_rgba = data_rgba.astype(np.float32)/255

    data_src = data_rgba[:, :, :3]
    alpha_src = data_rgba[:, :, 3]

    data_bkg = 1
    data_rgb = data_src*alpha_src + data_bkg*(1 - alpha_src)

    data_rgb = np.clip(data_rgb, 0, 1)*255
    data_rgb = np.round(data_rgb).astype(np.uint8)

    # Done
    return data_rgb



def data_url(data_comp, fmt):
    """Assemble compressed image data into URL data string
    """
    data_encode = base64.b64encode(data_comp)

    encoding = 'utf-8'
    template = 'data:image/{:s};charset={};base64,{:s}'

    # The decoding step here is necesary since we need to interpret byte data as text.
    # See this link for a nice explanation:
    # http://stackoverflow.com/questions/14010551/how-to-convert-between-bytes-and-strings-in-python-3
    result = template.format(fmt, encoding, data_encode.decode(encoding=encoding))

    return result



def iter_tiles(img, size):
    """Generator over image tiles
    """

    num_lines, num_samples = img.shape[:2]

    num_chunk_lines = int(num_lines/size)
    chunk_lines = int(np.round(num_lines/num_chunk_lines))
    chunk_lines -= chunk_lines % 2

    num_chunk_samples = int(num_samples/size)
    chunk_samples = int(np.round(num_samples/num_chunk_samples))
    chunk_samples -= chunk_samples % 2

    for j in range(num_chunk_lines):
        j0 = j*chunk_lines
        slice_lines = slice(j0, j0 + chunk_lines)
        for i in range(num_chunk_samples):
            i0 = i*chunk_samples
            slice_samples = slice(i0, i0 + chunk_samples)

            yield img[slice_lines, slice_samples], j0, i0


#------------------------------------------------

if __name__ == '__main__':
    pass
